# Gcode_Exporter.rb - Exports gcode .gcode file from SketchUp
#
# Copyright (C) 2013 David Fishelzon
#
# License: Apache License, Version 2.0
#
# G-Code Exporter
#
require 'sketchup'

module GcodeExporterModule  # Namespace for THIS plugin

  class GcodeExporter
    load File.join(File.dirname(__FILE__),'Gcode.ini')
      FilamentWidth = 3.mm
      NozelDiameter = 0.5.mm
      FilaCutArea                 = Math::PI * ( FilamentWidth*FilamentWidth / 4 ) 
      NozelDiameterDivFilaCutArea = NozelDiameter / FilaCutArea

=begin
      unless Gcode_Exporter_File_Heder 
             Gcode_Exporter_File_Heder  = ";Gcode Exporter File Heder"
      end
      unless Gcode_Exporter_Layer_Heder
             Gcode_Exporter_Layer_Heder = "; Gcode Exporter Layer %s Heder "
      end
      unless Gcode_Exporter_Layer_Futer
             Gcode_Exporter_Layer_Futer = "; Gcode Exporter Layer %s Futer "
      end
      unless Gcode_Exporter_File_Futer
             Gcode_Exporter_File_Futer  = ";Gcode Exporter File Futer"
      end
=end

    #
    # exportLayerGcode exports the gcode in a single layer
    #
    def exportLayerGcode startPoint
      old_x = old_y = old_z = old_f = old_e = ''
      currPoint = startPoint
      begin
        # Find the segment at currPoint
        @ptr = 0
        while @ptr < @entsInLayer.length
          if currPoint.vector_to(@entsInLayer[@ptr][0]).length.to_f < 0.01.mm
            # while - Found the next segment!
            currPoint = @entsInLayer.delete_at(@ptr)
            if currPoint[2] # component is a filament 
              compLength    = currPoint[0].vector_to(currPoint[1]).length.to_f
              compHight     = currPoint[4] 
              e = "E%0.5f" % ( compLength * compHight * NozelDiameterDivFilaCutArea )
            else            # component is a Head Move
              e = "E0"
            end # if component a filament or a Head move
            x = "X%0.4f" % (currPoint[1][0] / 1.mm)
            y = "Y%0.4f" % (currPoint[1][1] / 1.mm)
            z = "Z%0.4f" % (currPoint[1][2] / 1.mm)
            f = "F%0.4f" %  currPoint[3]
            xu = yu = zu = eu = fu = ''
            xu = x unless x == old_x
            yu = y unless y == old_y
            zu = z unless z == old_z
            fu = f unless f == old_f
            eu = e unless e == old_e
            old_x, old_y, old_z, old_f, old_e = x,y,z,f,e

            @gcodeFile.puts "G1 #{xu} #{yu} #{zu} #{fu} #{e}"

            currPoint = currPoint[1]
            @ptr = @entsInLayer.length + 1 # stop looking - move to end +1
          else # try next segment
            @ptr += 1
          end # if
        end # while @ptr < @entsInLayer.length
      end until @entsInLayer.length == 0 or @ptr == @entsInLayer.length
      return currPoint
    end #exportLayerGcode

    #
    # distance to the closest previos component
    #
    def closest_previos_component_to my_index
      my_startPo = @entsInLayer[ my_index ][0]
      min_dist   = 1000000.0.mm # to previos end
      @entsInLayer.each_with_index {|item, index|
        unless index == my_index
          dist = item[1].vector_to(my_startPo).length.mm.to_f 
          if min_dist > dist
             min_dist = dist
          end # if
        end #unless
      } # each_with_index 
      return min_dist
    end # def closest_previos_component_to

    #
    # find the first link of a list of components
    #
    def findFirstLink 
      # find for each component the "distance to the closest previos component"
      # the component with the biggest distance is the "first link"
      currPoint = @entsInLayer[0][0]
      return currPoint if @entsInLayer.length < 2
      max_dist  = 0
      @entsInLayer.each_with_index {|item, index|
        dist = self.closest_previos_component_to index
        if max_dist < dist
           max_dist = dist
          currPoint = item[0]
        end # if
      } #each_with_index 
      return currPoint.to_a
    end # def findFirstLink 

    #
    # exportGcode will sort all connected Gcode components by layers
    # 
    def exportGcode
      maxMoveSpeed = 7800.0
      @gcodeFile.puts Gcode_Exporter_File_Heder

      # Get Layar list and an array for each layer's components
      @layersList    = []
      @export_ents.map {|ent| 
        a_headMove   = nil
        entHight     = 0.0
        a_filament   = @filamentComp.instances.include? ent
        if a_filament
          edge       = ent.entities[12]
          entHight   = ent.bounds.depth / 1.mm #bounds.width=X, bounds.height=Y, bounds.depth=Z.
        end #  if a_filament
        unless a_filament # chack maybe ent is a head move
          a_headMove = @HeadMoveComp.instances.include? ent 
          edge       = ent.entities[0] if a_headMove
        end # unless a_filament
        if a_filament or a_headMove
          lay     = ent.layer.name
          layEval = "@lay_" + lay.split.join.downcase
          unless @layersList.include?( lay )
            @layersList << lay
            eval "#{layEval} = [] "
          end #  @layersList.include?( lay )
          trnsf = ent.transformation
          strPo = edge.start.position.transform!(trnsf)
          endPo = edge.end.position.transform!(trnsf)
          speed = ent.get_attribute "physical_properties", "Move_Speed"
          speed = maxMoveSpeed unless speed
          eval "#{layEval}  << [strPo.to_a, endPo.to_a, a_filament, speed, entHight] "
        end # if ent.is_a Component
      } # @export_ents.map
      @layersList.sort!

      point = Geom::Point3d.new(0, 0, 0)
      # export all layers
      @layersList.each  {|lay|
        layEval = "@lay_" + lay.split.join.downcase
        eval "@entsInLayer = #{layEval} "
        @gcodeFile.puts Gcode_Exporter_Layer_Heder % lay
        while @entsInLayer.length > 0
          point = self.exportLayerGcode point
          if @entsInLayer.length > 0
            # unconnected component found 
            @gcodeFile.puts "; #{@entsInLayer.length} Components are not connected at layer #{lay} - connecting them"
            # try to connect them find a first link of unconnected components 
            @entsInLayer << [point, self.findFirstLink, false, maxMoveSpeed, 0]
          end # if
        end # while
        @gcodeFile.puts Gcode_Exporter_Layer_Futer % lay
        Sketchup.status_text="finished exporting layer #{lay} "
      } # @layersList.each
      
      @gcodeFile.puts Gcode_Exporter_File_Futer
    end # def exportGcode

    def exportGcodeUI 
      @model        = Sketchup.active_model
      @filamentComp = @model.definitions[ "filament" ]
      @HeadMoveComp = @model.definitions[ "HeadMove" ]
      return nil   if not @filamentComp or not @HeadMoveComp 

      # Start Undo
      if (Sketchup.version.to_i < 7)
        @model.start_operation("Export G-Code")
      else
        @model.start_operation("Export G-Code",true)
      end

      if @model.selection.empty?
        @export_ents = @model.entities
      else
        @export_ents = @model.selection
      end

      if ( @export_ents.length > 0 )
        # Get exported file name and export.
        @out_name = File.basename(@model.path).split(".")[0]
        @out_name = @model.description.split(/[\\.]/)[-2] if not @out_name
        @out_name = "untitled"                            if not @out_name
        @out_name = UI.savepanel( "Gcode file location", "", "#{@out_name}.gcode" )
        if @out_name
          @gcodeFile = File.new( @out_name , "w" )  
          puts "~~~~Starting~~~~~"
          Sketchup.status_text="Gcode export Stated"
          self.exportGcode()

          puts "~~~~FINISH~~~~~"
          Sketchup.status_text=""
          @gcodeFile.close
          UI.messagebox( " Gcode file exported #{@out_name}" )
        end # @out_name exist
      end # if length > 0
     
      @model.commit_operation # End Undo
    end # def exportGcodeUI 

  end # class GcodeExporter

  # Register Menus in the UI
  unless file_loaded?(__FILE__) 

    UI.menu('Plugins').add_item("Fishelzon's G-Code Exporter...")     { 
      GcodeExporter.new.exportGcodeUI 
    } # UI.menu
=begin
    UI.menu('Plugins').add_item("Fishelzon's G-Code Exporter Reload") { 
      if load( __FILE__ ) 
        Sketchup.send_action 'showRubyPanel:'
      end #if load( __FILE__ ) 
    } # UI.menu Reload
=end
  end # unless file_loaded?
  file_loaded(__FILE__)

end # module GcodeExporterModule
